# eScriptorium/OpenITI : virtual keyboard

- Version 1: published on 31 May 2021
- Version 2: published on 2 June 2021
    - use UTF-8 scripts instead of blocks
    - support UTF-8 Private Use Areas
    - simpler keyboard grid size edition
    - added implementation steps
- Version 3: published on 3 June 2021
    - download keyboards file
    - mention dead-keys as a potential followup

## Summary

Some eScriptorium users need to input UTF-8 glyphs that are not present on physical keyboards, and generally hard to input.
The transcription editor would greatly benefit from having a virtual keyboard customizable by user.
Each user would have the ability to add virtual keys to the input text area, and categorize those keys by virtual keyboard. For example, someone could have:
 - a virtual keyboard with 10 glyphs to transcribe Ancient English
 - another virtual keyboard with 4 Cyrillic glyphs

Switching between virtual keyboards would be as simple as picking a choice in a list. The physical keyboard would still be fully usable, the virtual keyboard would be usable through mouse and keyboard shortcuts (Ctrl + numbers for example).

Virtual keyboards configuration would be stored as JSON, available in the browser local storage, but also shareable between users (through an import/export feature)

A configuration interface would allow a user to:
 - manage keyboards (creation, edition, deletion)
 - manage glyphs in a keyboard:
   - add glyphs
   - add a shortcut
   - change position in keyboards
 - export a keyboard configuration as a JSON file
 - import a keyboard configuration from a JSON file

This software would be developed as a standalone Vue.js project.
It could be used in two ways:
1. direct integration in the eScriptorium project.
This would need some development in eScriptorium itself, but the feature would be immediately available to all users.
2. as a cross-platform webextension (for Firefox and Chrome)
This should not require any change on eScriptorium, but users would have to install the extension themselves.

## Keyboard layout

We'll only support a simple grid layout (using rows and columns), as the virtual keyboard will not be used directly by typists and thus does not benefit from a classic representation with diagonal keys.

Blank keys are possible, as described below in the mockups.

## Screens

### Keyboard display

- A user should be able to select glyphs for a textarea/input on a website
- A user should be able to switch between keyboards

![](spec_keyboard.png)


Actions:
 - clicking on a glyph inputs the glyph on the current textarea
 - selecting a keyboard immediately displays that keyboard
 - clicking on **Manage keyboards** displays the keyboard management panel described below



### Keyboard management

- A user should be able to list all available keyboards
- A user should be able to access keys configuration for that keyboard
- A user should be able to delete a keyboard
- A user should be able to create a keyboard


![](spec_list.png)



Actions:
- clicking on the **keyboard name** brings the keyboard display of that keyboard
- clicking on **Edit** displays the configuration of that keyboard (described below)
- clicking on **Export** provide a JSON file for download
- clicking on **Delete** shows a confirmation modal (alert) asking if the user really wants to delete that keyboard. 
- clicking on **Create a new keyboard** adds immediately a new keyboard named `New Keyboard`



### Keyboard configuration

- A user should be able to rename a keyboard
- A user should be able to resize a keyboard
- A user should be able to position characters
- A user should be able to select Utf-8 glyphs on the keyboard
- A user should be able to assign a keypress to that glyph


![](spec_configure.png)

Actions:
- Clicking on **Save** updated the keyboard name
- Clicking on a keyboard cap will bring  character details on the right side
- Increasing the number of rows immediately add a new row at the bottom
- Decreasing the number of rows will remove the last row
    - if it's empty
- Increasing the number of columns immediately add a new column at the right
- Decreasing the number of columns will remove the last column
    - if it's empty

- Clicking on the **UTF-8 picker** will display the character selection screen described below
- update the value of a keypress to bind a physical keypress to the character
- input an utf-8 code to manually specify the code, instead of relying on the UTF-8 picker

Note: to remove a non-empty row or column, the user would have to **clear** all extra characters (one by one).

#### Character selection from all avaialble UTF-8 characters


![](spec_picker.png)

- select a script from all known ones
- display all characters in the script (one line of 16 characters to match hexa description)
- select a character, ask for validation (single button to use it)

![](spec_glyph.png)


Actions:
- Select a character block from a list
- Select a character from a block and view its name
- Use selected character it in the current keyboard
- Import a JSON file describing a new Private Use Area script (using the format described in last sections)


### Keyboard export
- A user should be able to export a keyboard as JSON

This action is available from the keyboard management panel (one click brings a download)

### Keyboard import

- A user should be able to import a keyboard from a JSON file
- A user should be able to import a keyboard from a JSON url


These actions are available from the keyboard management panel

Two forms would be available on a dedicated Vue.js component:
- one form with a single file upload field, allowing a user to upload a JSON file describing a keyboard. This would be used to import a single keyboard from a local file (sent by email for example)
- another form with a single URL input field, allowing a user to retrieve a remote JSON file describing a keyboard. This would be used to import a single keyboard from a remote file (hosted as a gist on github.com or simply as a file on a personal server)

Both forms would display any errors occuring during the retrieval and parsing of the keyboard specifications.

As soon as a form is imported in `localStorage`, it's available to users.


## Keyboard JSON structure

A JSON file with the following data. The same structure is used both in `localStorage` and to export a keyboard.


```json
{
    "version": "1.0.0",
    "name": "XXX",
    "author": "Bastien Abadie <abadie@teklia.com>"
    "characters": [
         {
             "row": 1,
             "column": 1,
             "character": "\u004F"
             "keyboard_code": 83
         }
    ]
}
```

- `row` & `column` are the position of the character on the virtual keyboard grid
- `character` is the hexadecimal representation of an utf-8 character
- `keyboard_code` is the numerical value of an optional **keydown** Javascript event that is used to type the character in the targeted field.
    - This allows users to bind a physical keyboard touch to a character from the virtual keyboard


## Vue.js implementation

### UTF-8 scripts

A set of static JSON files with known UTF-8 scripts will be stored in the application.
Each UTF-8 script will be stored as an independant JSON file with the following data:
- script name
- all characters in the script, ordered position, with metadata:
    - codepoint
    - description (known name)

```json
{
    "name": "Basic Latin",
    "characters": [
        {
            "code": "0041",
            "description": "LATIN CAPITAL LETTER A"
        },
        ...
    ]
}
```

This would enable us to also support [Private Use Areas](https://en.wikipedia.org/wiki/Private_Use_Areas) code points (like [MUFI](https://mufi.info))

### Store

A simple store reading/writing from `localStorage` will be implemented, using [vue-local-store](https://github.com/pinguinjkeke/vue-local-storage)

Data stored:
- Keyboards using the format described above
- Currently used keyboard
- Last used UTF-8 block when editing

### Components

| Component         | Description                                                 | Parent         |
|:----------------- |:----------------------------------------------------------- | -------------- |
| KeyboardDisplay   | Shows a keyboard and allow user to pick a character         |                |
| KeyboardManager   | List all available keyboards and possible actions           |                |
| KeyboardEditor    | Show current disposition of a keyboard                      |                |
| KeyboardCharacter | A specific character and its details                        | KeyboardEditor |
| KeyboardImport    | A form to upload and parse a JSON file to import a keyboard |                |
| UTF8Picker        | Displays UTF-8 scripts and their characters                 |                |
| UTF8Char          | A single UTF-8 character and its details                    | UTF8Picker     |
| UTF8ScriptCreator | Add a new UTF-8 script from a JSON Payload                              |                |

## Unicode

- [Official character charts](http://www.unicode.org/charts/)
- [Nicer way to browse Blocks](https://unicode-table.com/en/blocks/)
- [Raw list of char names from version 13.0](https://www.unicode.org/Public/13.0.0/ucd/extracted/DerivedName.txt)
- [Raw list of scripts](https://www.unicode.org/Public/13.0.0/ucd/Scripts.txt)
- [PUA Medieval MUFI](https://mufi.info/)


## WebExtension

Some technical limitations will require some changes for the webextension:
- A URL whitelist should be implemented to avoid triggering the virtual keyboard on most websites
- The webextension would only support HTML elements `textarea` and `input[type=text]`
- As a webpage can have multiple `input` and `textarea` elements, the webextension would only use the one in focus (after a click usually)
- The webextension would display the virtual keyboard just above the `input` or `textarea` targeted element
- All settings would be implemented on the extension settings webpage (not visible on a website)

## Implementation steps

Here is a rough summary of the steps we'll go through to start the implementation.

1. POCs for technical challenges
    - sharing codebase between JS lib & webextension
    - changing a keyboard input in a focused HTML texarea (when I press `A`, the textarea gets `B`)
2. Build UTF-8 scripts JSON payloads from online references
    - official scripts
    - MUFI
4. Build UTF-8 picker as Vue.js component
5. Build KeyboardEditor as Vue.js component
6. Build KeyboardManager as Vue.js component
7. Build KeyboardDisplay as Vue.js component
8. Build KeyboardImport as Vue.js component
9. Build UTF8ScriptCreator as Vue.js component
10. Write final documentation

At each steps, we'll produce both a JS library and the web extension implementation.

## Deliverables

- All source code in a public repository under the open-source licence of your choice
- Markdown documentation into the repository
- Simple HTML examples on how to embed the javascript build on a website
- One merge request on the eScriptorium project to embed the javascript build
- Publication on the Mozilla webextension store and the Google chrome web store

## Potential followups

### "Dead" keys

The goal of this feature would be to support a sub-keymap for some keys.

> A simple example would be Ethiopic where you might click on a letter to select the consonant, then get subkeymap with each order of that consonant (the attached vowel).

It's a feature implemented on [Pinyin](https://en.wikipedia.org/wiki/Pinyin_input_method) keyboards or virtual input systems to be able to select a glyph across a large range of characters.

The project would need to support said sub-keymaps both in the keyboard editor and in the keyboard file structure.
